/* THIS FILE IS CREATED AUTOMATICALLY - DO NOT EDIT */

#define	PROTO_OSPF	1
#define	PROTO_ASPATHS	1
#define	GDC_RESOURCE	1
#define	IFL_NAMELEN	16
#define	U_INT16	unsigned short
#define	VOID_T	void *
#define	SIGRETURN	return
#define	KRT_RTREAD_KMEM	1
#define	PROTO_INET	1
#define	POSIX_SIGNALS	1
#define	HAVE_DIRENT	1
#define	KSYM_IPFORWARDING	"_ipforwarding"
#define	KRT_RT_IOCTL	1
#define	PROTO_ICMP	1
#define	HAVE_WAITPID	1
#define	INCLUDE_UNISTD	1
#define	PROTO_SNMP	1
#define	KVM_TYPE_SUNOS4	1
#define	SIGTYPE	void
#define	NEED_STRERROR	1
#define	AUTONOMOUS_SYSTEM	1
#define	PROTO_HELLO	1
#define	PROTO_ISODE_SNMP	1
#define	KRT_SYMBOLS_NLIST	1
#define	PROTO_BGP	1
#define	LOG_FACILITY	LOG_DAEMON
#define	ROUTER_ID	1
#define	S_INT32	int
#define	PID_T	pid_t
#define	KRT_LLADDR_SUNOS4	1
#define	KRT_IFREAD_IOCTL	1
#define	PROTO_EGP	1
#define	DEBUG	1
#define	U_INT8	unsigned char
#define	MIB_RIP	1
#define	RT_N_MULTIPATH	4 
#define	INET_CLASS_E	1
#define	__GATED__	1
#define	S_INT16	short
#define	NEED_SIGNAME	1
#define	ENVIRON	environ
#define	FLETCHER_CHECKSUM	1
#define	PROTO_RIP	1
#define	KSYM_UDPCKSUM	"_udpcksum"
#define	GID_T	int
#define	NEED_INET_ATON	1
#define	LOG_OPTIONS	LOG_PID | LOG_CONS | LOG_NDELAY
#define	U_INT32	unsigned int
#define	PARSE_PORT	1
#define	S_INT8	char
