/*
 *  isis_flood.h,v 1.11 1993/01/07 22:39:18 jch Exp
 */

/* Gated Release 3.5 */
/* Copyright (c) 1990,1991,1992,1993,1994 by Cornell University. All rights */
/* reserved. Refer to Particulars and other Copyright notices at the end of this */
/* file. */


#ifndef _flood_h
#define _flood_h

/*
 *	Quick summary of critical information about the LSP
 */
typedef struct _LSSignature {
	LSPID		id;				/* ID of LSP originator */
	u_long		seqNum;
	u_short		lifetime;		/* reasonably up-to-date remaining time */
	int			level;			/* 1 or 2 */
	u_short		checksum;
} LSSignature;

/*
 *	Structure to hold each LSP in the database.
 */
typedef struct _LSEntry {
	DLList		links;				/* link: timeout list */
	LSSignature	signature;
	u_short		originalLifetime;		/* value of lifetime when received */
	u_short		remainingTime;			/* time to live used in ageList */
	long		arrivalTimestamp;		/* when pdu arrived/updated */
	Boolean		SRM[MaximumCircuits];
	Boolean		SSN[MaximumCircuits];
	Boolean		foundInSNP;		/* used for processing CSNPs */
	Boolean		onDeletedList;		/* true if marked for deletion */
	u_char		*buf;			/* this baby is task_block_alloc'd */
	int		allocLen;		/* the actual amount alloc'd (different from len) */
	int		len;			/* amount of buffer space this LSP is using (different from allocLen) */
	struct _LSPDesc	*lspDesc;		/* ptr to LSPDesc, if I build this lsp */
	struct _LSEntry	*hashNext;		/* next in hash table */
	struct _LSEntry	*sortNext;		/* next in sorted list */
} LSEntry;

#define locallyGenerated(ent) (ent->lspDesc != NULL)

/*
 * This guy holds a list of LS entries containing only prefixes
 * in the case that no other entries changed (i.e., no interior
 * nodes in the SPF graph have changed, so I just need to adjust
 * routes for the prefixes contained in this list).
 */
struct _lseChangeList {
	struct _lseChangeList *next;
	struct _LSEntry *lse;
};

/*
 *	Various flags to the SRM/SSN routines
 */
#define	AllCircuits		 0x01
#define	SingleCircuit	 0x02

#endif 	/* _flood_h */


/*
 * ------------------------------------------------------------------------
 * 
 * 	GateD, Release 3.5
 * 
 * 	Copyright (c) 1990,1991,1992,1993,1994 by Cornell University
 * 	    All rights reserved.
 * 
 * 	THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY
 * 	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
 * 	LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * 	AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * 	Royalty-free licenses to redistribute GateD Release
 * 	3 in whole or in part may be obtained by writing to:
 * 
 * 	    GateDaemon Project
 * 	    Information Technologies/Network Resources
 * 	    200 CCC
 * 	    Cornell University
 * 	    Ithaca, NY  14853-2601  USA
 * 
 * 	GateD is based on Kirton's EGP, UC Berkeley's routing
 * 	daemon	 (routed), and DCN's HELLO routing Protocol.
 * 	Development of GateD has been supported in part by the
 * 	National Science Foundation.
 * 
 * 	Please forward bug fixes, enhancements and questions to the
 * 	gated mailing list: gated-people@gated.cornell.edu.
 * 
 * ------------------------------------------------------------------------
 * 
 *       Portions of this software may fall under the following
 *       copyrights:
 * 
 * 	Copyright (c) 1988 Regents of the University of California.
 * 	All rights reserved.
 * 
 * 	Redistribution and use in source and binary forms are
 * 	permitted provided that the above copyright notice and
 * 	this paragraph are duplicated in all such forms and that
 * 	any documentation, advertising materials, and other
 * 	materials related to such distribution and use
 * 	acknowledge that the software was developed by the
 * 	University of California, Berkeley.  The name of the
 * 	University may not be used to endorse or promote
 * 	products derived from this software without specific
 * 	prior written permission.  THIS SOFTWARE IS PROVIDED
 * 	``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * 	INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * 	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * ------------------------------------------------------------------------
 * 
 * 	Copyright 1991 D.L.S. Associates
 * 
 * 	Permission to use, copy, modify, distribute, and sell this software
 * 	and its documentation for any purpose is hereby granted without
 * 	fee, provided that the above copyright notice appear in all copies
 * 	and that both that copyright notice and this permission notice
 * 	appear in supporting documentation, and that the name of D.L.S. not
 * 	be used in advertising or publicity pertaining to distribution of
 * 	the software without specific, written prior permission.  D.L.S.
 * 	makes no representations about the suitability of this software for
 * 	any purpose.  It is provided "as is" without express or implied
 * 	warranty.
 * 
 * 	D.L.S. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * 	NO EVENT SHALL D.L.S.  BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * 	OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * 	NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * 	CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * 	Authors:  Robert Hagens and Dan Schuh
 * 
 * 
 */
