/*
 * icmp.h,v 1.10 1993/10/27 01:39:47 jch Exp
 */

/* Gated Release 3.5 */
/* Copyright (c) 1990,1991,1992,1993,1994 by Cornell University. All rights */
/* reserved. Refer to Particulars and other Copyright notices at the end of this */
/* file. */


#ifdef	PROTO_ICMP

#define	ICMP_PACKET_MAX	2048	/* Maximum ICMP packet size we expect */

extern struct _task *icmp_task;
extern trace *icmp_trace_options;
extern const bits icmp_trace_types[];

/* Tracing */

#define	TR_ICMP_DETAIL_REDIRECT	TR_DETAIL_1
#define	TR_ICMP_DETAIL_ROUTER	TR_DETAIL_2
#define	TR_ICMP_DETAIL_INFO	TR_DETAIL_3
#define	TR_ICMP_DETAIL_ERROR	TR_DETAIL_4

#define	TR_ICMP_INDEX_PACKETS	0
#define	TR_ICMP_INDEX_REDIRECT	1
#define	TR_ICMP_INDEX_ROUTER	2
#define	TR_ICMP_INDEX_INFO	3
#define	TR_ICMP_INDEX_ERROR	4

   
PROTOTYPE(icmp_var_init,
	  extern void,
	  (void));
PROTOTYPE(icmp_init,
	  extern void,
	  (void));

#endif	/* PROTO_ICMP */


/*
 * ------------------------------------------------------------------------
 * 
 * 	GateD, Release 3.5
 * 
 * 	Copyright (c) 1990,1991,1992,1993,1994 by Cornell University
 * 	    All rights reserved.
 * 
 * 	THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY
 * 	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
 * 	LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * 	AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * 	Royalty-free licenses to redistribute GateD Release
 * 	3 in whole or in part may be obtained by writing to:
 * 
 * 	    GateDaemon Project
 * 	    Information Technologies/Network Resources
 * 	    200 CCC
 * 	    Cornell University
 * 	    Ithaca, NY  14853-2601  USA
 * 
 * 	GateD is based on Kirton's EGP, UC Berkeley's routing
 * 	daemon	 (routed), and DCN's HELLO routing Protocol.
 * 	Development of GateD has been supported in part by the
 * 	National Science Foundation.
 * 
 * 	Please forward bug fixes, enhancements and questions to the
 * 	gated mailing list: gated-people@gated.cornell.edu.
 * 
 * ------------------------------------------------------------------------
 * 
 *       Portions of this software may fall under the following
 *       copyrights:
 * 
 * 	Copyright (c) 1988 Regents of the University of California.
 * 	All rights reserved.
 * 
 * 	Redistribution and use in source and binary forms are
 * 	permitted provided that the above copyright notice and
 * 	this paragraph are duplicated in all such forms and that
 * 	any documentation, advertising materials, and other
 * 	materials related to such distribution and use
 * 	acknowledge that the software was developed by the
 * 	University of California, Berkeley.  The name of the
 * 	University may not be used to endorse or promote
 * 	products derived from this software without specific
 * 	prior written permission.  THIS SOFTWARE IS PROVIDED
 * 	``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * 	INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * 	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
